/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) 1989-2004 MAXON Computer GmbH, all rights reserved  //
/////////////////////////////////////////////////////////////

// volumetric shader example that accesses particles and displays its own preview
// absolutely not optimized for speed

#include "c4d.h"
#include "c4d_symbols.h"
#include "customgui_matpreview.h"

struct PVRender
{
	PVRender(void);
	~PVRender(void);

	Vector	*mp;
	LONG		count;
};

class ParticleVolume : public MaterialData
{
	private:
		PVRender *render;
	public:
		virtual	VOLUMEINFO GetRenderInfo		(BaseMaterial *mat);

		virtual	INITRENDERRESULT InitRender				(BaseMaterial *mat, const InitRenderStruct &irs);
		virtual	void FreeRender				(BaseMaterial *mat);

		virtual	void CalcSurface			(BaseMaterial *mat, VolumeData *vd);
		virtual	void CalcTransparency	(BaseMaterial *mat, VolumeData *vd);
		virtual	void CalcVolumetric		(BaseMaterial *mat, VolumeData *vd);
    virtual Bool Init(GeListNode *node);
    virtual Bool Message(GeListNode *node, LONG type, void *data);

		static NodeData *Alloc(void) { return gNew ParticleVolume; }
  protected:
    Bool highDensity;
};

Bool ParticleVolume::Init(GeListNode *node)
{
  highDensity = FALSE;
  return TRUE;
}

VOLUMEINFO ParticleVolume::GetRenderInfo(BaseMaterial *mat)
{
	return VOLUMEINFO_TRANSPARENCY|VOLUMEINFO_VOLUMETRIC;
}

#define MAX_PARTICLES 10000

PVRender::PVRender(void)
{
	mp=NULL;
	count=0;
}

PVRender::~PVRender(void)
{
	GeFree(mp);
}

static Bool FillPV(PVRender *pv, BaseObject *op)
{
	while (op)
	{
		if (op->GetType()==Oparticle && !op->GetDown() && op->GetDeformMode()) // particle system without geometry
		{
			ParticleObject *po=(ParticleObject*)op;
			ParticleTag    *pt=(ParticleTag*)op->GetTag(Tparticle);
			if (pt)
			{
				Vector	 *mp  = NULL;
				const Particle *pp  = NULL;
				LONG		 pcnt = po->GetParticleCount(),rcnt=0,i;

				for (i=0; i<pcnt; i++)
				{
					pp=po->GetParticleR(pt,i);
					if (pp->bits&(PARTICLEFLAGS_VISIBLE|PARTICLEFLAGS_ALIVE)) rcnt++;
				}

				if (pv->count+rcnt>0)
				{
					mp = GeAllocType(Vector,pv->count+rcnt);
					if (!mp) return FALSE;

					CopyMemType(pv->mp, mp, pv->count );
					GeFree(pv->mp);
					pv->mp=mp;

					for (i=0; i<pcnt; i++)
					{
						pp=po->GetParticleR(pt,i);
						if ((pp->bits&(PARTICLEFLAGS_VISIBLE|PARTICLEFLAGS_ALIVE))==(PARTICLEFLAGS_VISIBLE|PARTICLEFLAGS_ALIVE)) 
							mp[pv->count++]=pp->off;
					}
				}
			}
		}

		if (!FillPV(pv,op->GetDown())) return FALSE;
		op=op->GetNext();
	}
	return TRUE;
}

INITRENDERRESULT ParticleVolume::InitRender(BaseMaterial *mat, const InitRenderStruct &irs)
{
	render = gNew PVRender;
	if (!render) return INITRENDERRESULT_OUTOFMEMORY;

	if (irs.doc && !FillPV(render,irs.doc->GetFirstObject()))
	{
		gDelete(render);
		return INITRENDERRESULT_OUTOFMEMORY;
	}

	return INITRENDERRESULT_OK;
}

void ParticleVolume::FreeRender(BaseMaterial *mat)
{
	gDelete(render);
}

Bool SphereIntersection(const Vector &mp, Real rad, Ray *ray, Real maxdist, Real *length)
{
	Vector	z;
	Real		kr,ee,det,s,s1;

	kr  = rad*rad;
	z   = ray->p.ToRV() - mp;
	ee  = ray->v.ToRV()*z;
	det = ee * ee - z*z + kr;
	
	if (det<0.0) return FALSE;

	det = Sqrt(det);
	s   = det - ee;

	if (s<=0.0) return FALSE;

	s1 = -det - ee;

	if (s1>=maxdist) return FALSE;

	if (s>=maxdist) s=maxdist;	
	if (s1<0.0) s1 = 0.0;

	*length = s-s1;
	
	return TRUE;
}

void ParticleVolume::CalcSurface(BaseMaterial *mat, VolumeData *vd)
{
	vd->trans = 1.0;
}

static Real GetCoverage(PVRender *pv, VolumeData *sd)
{
	#define PT_RADIUS 50.0

	LONG		 i;
	Real		 len,maxlen=0.0;

	for (i=0; i<pv->count; i++)			
	{
		if (!SphereIntersection(pv->mp[i],PT_RADIUS,sd->ray,sd->dist,&len)) continue;

		len =len/(2.0*PT_RADIUS);
		len*=len;
		len*=len;

		maxlen=len+maxlen-len*maxlen;
	}

	if (maxlen>1.0) maxlen=1.0;
	return maxlen;
}

void ParticleVolume::CalcVolumetric(BaseMaterial *mat, VolumeData *vd)
{
	Real cov=GetCoverage(render,vd);
	vd->col=Vector(0.0,0.0,cov);
	vd->trans=1.0-cov;
}

void ParticleVolume::CalcTransparency(BaseMaterial *mat, VolumeData *vd)
{
	Real cov=GetCoverage(render,vd);
	vd->trans=1.0-cov;
}

Bool ParticleVolume::Message(GeListNode *node, LONG type, void *data)
{
  switch (type)
  {
  case MATPREVIEW_GET_OBJECT_INFO:
    // the preview needs our object information
    {
      MatPreviewObjectInfo* info = (MatPreviewObjectInfo*)data;
      info->bHandlePreview = TRUE; // own preview handling
      info->bNeedsOwnScene = TRUE; // we need our own entry in the preview scene cache
      info->bNoStandardScene = TRUE; // create our own preview scene
      info->lFlags = MATPREVIEW_FLAG_HIDE_SCENE_SETTINGS; // hide the scene settings entry
      return TRUE;
    }
    break;
  case MATPREVIEW_MODIFY_CACHE_SCENE:
    // modify the preview scene here. We have a pointer to a scene inside the preview scene cache.
    // We set bNoStandardScene = TRUE, so the passed scene is an empty document
    {
      MatPreviewModifyCacheScene* scene = (MatPreviewModifyCacheScene*)data;
      // add an environment object and create a texture tag
      BaseObject* env = BaseObject::Alloc(Oenvironment);
      if (!env) return FALSE;
      env->SetName("environment");
      BaseTag* mattag = env->MakeTag(Ttexture);
      if (!mattag) return FALSE;
      scene->pDoc->InsertObject(env, NULL, NULL);
      // add a particle emitter
      BaseObject* emitter = BaseObject::Alloc(Oparticle);
      if (!emitter) return FALSE;
      emitter->SetName("emitter");
      scene->pDoc->InsertObject(emitter, NULL, NULL);
      // get the current camera and change the position
      BaseObject* cam = scene->pDoc->GetActiveBaseDraw() ? scene->pDoc->GetActiveBaseDraw()->GetEditorCamera() : NULL;
      if (cam)
			{
				cam->SetRelPos(Vector(0, 0, -150));
				cam->SetRelRot(Vector(0, 0, 0));
			}
      return TRUE;
    }
    break;
  case MATPREVIEW_PREPARE_SCENE:
    {
      // this is called each time the preview wants to render our scene
      MatPreviewPrepareScene* preparescene = (MatPreviewPrepareScene*)data;

      AutoAlloc<AliasTrans> trans;
      if (!trans)
        return FALSE;
      if (!trans->Init(GetActiveDocument()))
        return FALSE;
      BaseMaterial* matclone = (BaseMaterial*)(Get()->GetClone(COPYFLAGS_0, trans));
      preparescene->pDoc->InsertMaterial(matclone);
      trans->Translate(TRUE);
      if (preparescene->pLink) preparescene->pLink->SetLink(matclone); // necessary

      // find the environment object and set the material
      BaseObject* env = preparescene->pDoc->SearchObject("environment");
      if (env)
      {
        TextureTag* textag = (TextureTag*)env->GetTag(Ttexture);
        if (textag)
          textag->SetMaterial(matclone);
      }

      // change the emitter data depending on the highDensity settings
      BaseObject* emitter = preparescene->pDoc->SearchObject("emitter");
      if (emitter)
      {
        BaseContainer* data = emitter->GetDataInstance();
        data->SetLong(PARTICLEOBJECT_BIRTHRAYTRACER, highDensity ? 10 : 5);
      }

      preparescene->bScenePrepared = TRUE; // inform the preview that the scene is prepared now
      return TRUE;
    }
    break;
  case MATPREVIEW_GET_POPUP_OPTIONS:
    {
      BaseContainer* bc = (BaseContainer*)data;
      bc->SetString(MATPREVIEW_POPUP_NAME, GeLoadString(IDS_PARTICLEVOLUME)); // element with index MATPREVIEW_POPUP_NAME is the text of the popup item
      bc->SetString(1, GeLoadString(IDS_PARTICLEVOLUME_HIGH_DENSITY) + (highDensity ? "&c&" : "")); // container entries should start with 1
      bc->SetString(0, ""); // add an example separator
      bc->SetString(2, GeLoadString(IDS_PARTICLEVOLUME_LOW_DENSITY) + (!highDensity ? "&c&" : ""));
      return TRUE;
    }
    break;
  case MATPREVIEW_HANDLE_POPUP_MSG:
    {
      LONG l = *((LONG*)data);
      if (l == 1) highDensity = TRUE;
      else if (l == 2) highDensity = FALSE;
      return TRUE;
    }
    break;
  case MATPREVIEW_GENERATE_IMAGE:
    {
      MatPreviewGenerateImage* image = (MatPreviewGenerateImage*)data;

      if (image->pDoc)
      {
        LONG w = image->pDest->GetBw();
        LONG h = image->pDest->GetBh();
        BaseContainer bcRender = image->pDoc->GetActiveRenderData()->GetData();
        bcRender.SetReal(RDATA_XRES, w);
        bcRender.SetReal(RDATA_YRES, h);
        bcRender.SetLong(RDATA_ANTIALIASING, ANTI_GEOMETRY);

        image->pDest->Clear(0, 0, 0);
        BaseTime bt(image->rTime + 1.0); // add one second that some particles are produced
        image->pDoc->SetTime(bt); 
        image->pDoc->ExecutePasses(image->pThread, TRUE, TRUE, FALSE, BUILDFLAGS_0);
        image->lResult = RenderDocument(image->pDoc, bcRender, NULL, NULL, image->pDest, 
        RENDERFLAGS_EXTERNAL | RENDERFLAGS_PREVIEWRENDER, image->pThread);
      }
      return TRUE;
    }
    break;
  }
  return MaterialData::Message(node, type, data);
}


// be sure to use a unique ID obtained from www.plugincafe.com
#define ID_PARTICLEVOLUME	1001163

Bool RegisterParticleVolume(void)
{
  String name = GeGetDefaultFilename(DEFAULTFILENAME_SHADER_VOLUME)+GeLoadString(IDS_PARTICLEVOLUME); // place in default Shader section
	return RegisterMaterialPlugin(ID_PARTICLEVOLUME,name,0,ParticleVolume::Alloc,"MParticleVolume",0);
}
