/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) MAXON Computer GmbH, all rights reserved            //
/////////////////////////////////////////////////////////////

#ifndef _LIB_SND_H_
#define _LIB_SND_H_

#include "c4d_library.h"
#include "c4d_baseplugin.h"

struct GeSndInfoEx;	// deprecated

#ifndef _INTERNAL_DEF_SND_

class BaseSound;

class GeSndInfo
{
private:
	LONG	channel_cnt;
	LONG	bits_per_channel;
	Real	sample_rate;
	LONG	sample_cnt;  // number of samples per channel
	CHAR	*data;

public:
	GeSndInfo(void)
	{
		channel_cnt = 0;
		bits_per_channel = 0;
		sample_rate = 0.0;
		sample_cnt = 0;
		data = NULL;
	}
  GeSndInfo(BaseSound *bs);

	// return the number of channels
	LONG	GetChannelCount( void ) const
	{
		return channel_cnt;
	}

	// return the number of bits per channel
	LONG	GetBitsPerChannel( void ) const
	{
		return bits_per_channel;
	}

	// return the sample rate in Hz
	Real GetSampleRate( void )
	{
		return sample_rate;
	}

	// return the duration/length of the sound in samples
	LONG	GetSampleCount( void ) const
	{
		return sample_cnt;
	}
	
	// return the length as BaseTime
	BaseTime	GetLength( void )
	{
		return BaseTime( GetSampleRate() != 0.0  ? (Real) sample_cnt / GetSampleRate() : 0.0 );
	}
};

class GePlaySnd
{

	private:
		GePlaySnd();
		~GePlaySnd();

	public:
		static GePlaySnd*		Alloc();
		static void					Free(GePlaySnd *&data);

		//----------------------------------------------------------------------------------------
		// Open audio device with a single BaseSound
		// The audio data referenced by the BaseSound will be kept in memory until you call Close()
		// Function result:		TRUE if successful
		// snd:								BaseSound
		//----------------------------------------------------------------------------------------
		Bool	Open( const BaseSound *snd );

		//----------------------------------------------------------------------------------------
		// Open audio device with multiple BaseSounds that will be mixed
		// The audio data referenced by the BaseSounds will be kept in memory until you call Close()
		// Function result:		TRUE if successful
		// stream_cnt:				number of BaseSounds to be mixed
		// streams:						array with BaseSound pointers
		// offsets:						array with start time offsets in seconds (this is the audible sound data when StartAt( 0 ) is called)
		// max_length:				maximum length of the audio data in seconds (relative to StartAt( 0 ))
		//----------------------------------------------------------------------------------------
		Bool	Open(  LONG stream_cnt, const BaseSound **streams, const Real *start_offsets, Real max_length );

		//----------------------------------------------------------------------------------------
		// Close audio device
		// If the BaseSounds used for Open() have been deleted in the meantime, their audio data will be released
		// Function result:		-
		//----------------------------------------------------------------------------------------
		void	Close( void );

		//----------------------------------------------------------------------------------------
		// Start audio output
		// Function result:		-
		// start:							position relative to the start of the audio data in seconds
		//----------------------------------------------------------------------------------------
		void	StartAt( LReal start );

		//----------------------------------------------------------------------------------------
		// Start audio scrubbing (this will automatically stop audio device after duration)
		// Function result:		-
		// start:							position relative to the start of the audio data in seconds
		// duration:					scrub duration in seconds after which the audio device will be stopped
		//----------------------------------------------------------------------------------------
		void	Scrub( LReal start, LReal duration );

		//----------------------------------------------------------------------------------------
		// Stop audio output
		// Function result:		-
		//----------------------------------------------------------------------------------------
		void	Stop( void );

		//----------------------------------------------------------------------------------------
		// Check if the audio device is running
		// Function result:		TRUE if audio is playing
		//----------------------------------------------------------------------------------------
		Bool	IsPlaying( void );
		
		//----------------------------------------------------------------------------------------
		// Return an estimate audio position
		// Function result:		the currently audible position in seconds
		//----------------------------------------------------------------------------------------
		LReal	GetPositionEstimate( void );

		//----------------------------------------------------------------------------------------
		// Set playback time scale
		// Function result:		-
		// scale							2.0 means twice as fast, 0.5 means half the speed, 0.0 is invalid
		//----------------------------------------------------------------------------------------
		void	SetScale( LReal scale );

		//----------------------------------------------------------------------------------------
		// Get playback time scale
		// Function result:		current playback time scale (1.0 means real time)
		//----------------------------------------------------------------------------------------
		LReal	GetScale( void );

};



#else

#define GePlaySnd		GeSndOut

#endif

// INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF
// INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF
// INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF


#define LIB_SND 1000496

struct GeSndLib : public C4DLibrary
{
		GePlaySnd*	(*Alloc)									();
		void				(*Free)										(GePlaySnd *&data);
		
		// dummy audio methods
		Bool				(GePlaySnd::*OpenEx)			(const GeSndInfoEx &si);
		void				(GePlaySnd::*CloseEx)			();
		void				(GePlaySnd::*StartEx)			(LONG start, LONG count);
		void				(GePlaySnd::*StopEx)			();
		Bool				(GePlaySnd::*IsPlayingEx)	();
		LONG				(GePlaySnd::*GetPositionEx)();

		// new audio stuff
		Bool				(GePlaySnd::*Open)				( LONG stream_cnt, const BaseSound **streams, const Real *start_offsets, Real max_length );
		void				(GePlaySnd::*Close)				();
		void				(GePlaySnd::*StartAt)			( LReal start );
		void				(GePlaySnd::*Scrub)				( LReal start, LReal duration );
		void				(GePlaySnd::*Stop)				();
		Bool				(GePlaySnd::*IsPlaying)		();
		LReal				(GePlaySnd::*GetPositionEstimate)	();

		void				(GePlaySnd::*SetScale)		( LReal scale );
		LReal				(GePlaySnd::*GetScale)		( void );
};																												
																													
// INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF
// INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF
// INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF -- INTERNAL STUFF

#endif
