#ifndef C4DMISC_DEFAULTALLOCATOR_H__
#define C4DMISC_DEFAULTALLOCATOR_H__

#include "../general.h"
#include <stdlib.h>
#include "c4d_memory.h"

namespace c4d_misc {

//----------------------------------------------------------------------------------------
/// Default implementation of an allocator.
/// An allocator is used by arrays, lists and other data structurs to allocate and release
/// memory. By default this implementation of an allocator is used.
///
/// There might be rare cases when you need a special memory alignment, a different resize
/// strategy or have to use a special memory area (stack, shared memory, ...). This can be
/// done by writing a custom allocator and specifying it as parameter upon array construction.
/// A custom allocator must implement the ComputeArraySize(), Alloc(), AllocClear(), Realloc()
/// and Free() methods, but it doesn't (and usually shouldn't) inherit from DefaultAllocator.
/// The allocator methods don't have to be static if your allocator requires member
/// variables, but the DefaultAllocator doesn't and therefore uses static methods for
/// better performance.
///
/// Please note that an allocator is copied upon array construction - it would be
/// a bad idea if your custom allocator object would consist of more than a few variables.
///
/// THREADSAFE
//----------------------------------------------------------------------------------------
class DefaultAllocator
{
public:
	//----------------------------------------------------------------------------------------
	/// Computes the new size for a growing array
	/// @return											new number of elements
	/// @param[in] current_size			current number of elements
	/// @param[in] increment				number of elements to be added (>= 1)
	/// @param[in] min_chunk_size		the minimum number of elements upon array creation
	/// THREADSAFE
	//----------------------------------------------------------------------------------------
	static Int ComputeArraySize(Int current_size, Int increment, Int min_chunk_size)
	{
		Int new_size	= (current_size + increment) * 3 / 2;	// increase required size by 50 % for future allocations
		Int remainder = new_size % min_chunk_size;
		if (remainder || new_size == 0)
			new_size += min_chunk_size - remainder;

		return new_size;
	}

	//----------------------------------------------------------------------------------------
	/// Allocates a memory block.
	/// The memory is not cleared, it may contain a certain byte pattern in debug mode
	/// @return									memory block address or nullptr
	/// @param[in] s						block size in bytes (values < 0 will return nullptr)
	/// THREADSAFE
	//----------------------------------------------------------------------------------------
	static inline void* Alloc(Int32 s, C4D_MISC_ALLOC_LOCATION_DECLARATION)
	{
		return _GeAllocNC((VLONG)s, (LONG)C4D_MISC_ALLOC_LOCATION_FORWARD.GetLine(), C4D_MISC_ALLOC_LOCATION_FORWARD.GetFile());
	}

	//----------------------------------------------------------------------------------------
	/// Allocates a memory block.
	/// The memory is not cleared, it may contain a certain byte pattern in debug mode
	/// @return									memory block address or nullptr
	/// @param[in] s						block size in bytes (values < 0 will return nullptr)
	/// THREADSAFE
	//----------------------------------------------------------------------------------------
	static inline void* Alloc(Int64 s, C4D_MISC_ALLOC_LOCATION_DECLARATION)
	{
		if (s!=Int(s)) { DebugStop(); return nullptr; }
		return _GeAllocNC((VLONG)s, (LONG)C4D_MISC_ALLOC_LOCATION_FORWARD.GetLine(), C4D_MISC_ALLOC_LOCATION_FORWARD.GetFile());
	}

	//----------------------------------------------------------------------------------------
	/// Allocates a memory block and clears it
	/// @return									memory block address or nullptr
	/// @param[in] s						block size in bytes (values < 0 will return nullptr)
	/// THREADSAFE
	//----------------------------------------------------------------------------------------
	static inline void* AllocClear(Int32 s, C4D_MISC_ALLOC_LOCATION_DECLARATION)
	{
		return _GeAlloc((VLONG)s, (LONG)C4D_MISC_ALLOC_LOCATION_FORWARD.GetLine(), C4D_MISC_ALLOC_LOCATION_FORWARD.GetFile());
	}

	//----------------------------------------------------------------------------------------
	/// Allocates a memory block and clears it
	/// @return									memory block address or nullptr
	/// @param[in] s						block size in bytes (values < 0 will return nullptr)
	/// THREADSAFE
	//----------------------------------------------------------------------------------------
	static inline void* AllocClear(Int64 s, C4D_MISC_ALLOC_LOCATION_DECLARATION)
	{
		if (s!=Int(s)) { DebugStop(); return nullptr; }
		return _GeAlloc((VLONG)s, (LONG)C4D_MISC_ALLOC_LOCATION_FORWARD.GetLine(), C4D_MISC_ALLOC_LOCATION_FORWARD.GetFile());
	}

	//----------------------------------------------------------------------------------------
	/// Resizes a memory block.
	/// The additional memory is not cleared, it may contain a certain byte pattern in debug mode
	/// @return									memory block address or nullptr if resize is not possible (p is still valid in this case)
	/// @param[in] p						current memory block (can be nullptr)
	/// @param[in] n						new block size in bytes (values < 0 will return nullptr)
	/// THREADSAFE
	//----------------------------------------------------------------------------------------
	static inline void* Realloc(void* p, Int n, C4D_MISC_ALLOC_LOCATION_DECLARATION)
	{
		return _GeReallocNC(p, (VLONG)n, (LONG)C4D_MISC_ALLOC_LOCATION_FORWARD.GetLine(), C4D_MISC_ALLOC_LOCATION_FORWARD.GetFile());
	}

	//----------------------------------------------------------------------------------------
	/// Frees a memory block
	/// @return									-
	/// @param[in,out] p				memory block address (can be nullptr, will be nullptr after return)
	/// THREADSAFE
	//----------------------------------------------------------------------------------------
	template<class T> static inline void  Free(T*& p)
	{
		void*	tmp = (void*) p;
		_GeFree(&tmp);
		p = nullptr;
	}
};

//----------------------------------------------------------------------------------------
/// Creates an object with constructor parameters (using new or T::Alloc, depending on the type). This does not throw any kind of
/// exception, it returns null if the allocation fails.
/// Usage: MyType* x = NewObj(MyType, optional constructor parameters);
/// @return									pointer to object (null if allocation failed)
/// @param[in] T						type of object
/// @param[in] ...					additional constructor parameters
/// THREADSAFE
//----------------------------------------------------------------------------------------
// The placement new operator does not throw (is noexcept). Therefore placement new has to make
// a null check before it calls the constructor and there is no need us to check the return value
// of DefaultAllocator::Alloc(). All compilers behave as expected (gcc, llvm, msvc and icc) and
// this according to the standard (§3.7.4.1).
#define NewObj(T, ...)  new (DefaultAllocator::Alloc(SIZEOF(T), C4D_MISC_ALLOC_LOCATION)) T(__VA_ARGS__)

//----------------------------------------------------------------------------------------
/// Deletes an object. This calls the destructor and frees memory afterwards.
/// @param[in,out] o				object pointer (can be nullptr, will be nullptr after return)
/// THREADSAFE
//----------------------------------------------------------------------------------------
template<typename T> void DeleteObj(T*& o)
{
	if (o)
	{
		o->~T();
		DefaultAllocator::Free(o);
		o = nullptr;
	}
}

//----------------------------------------------------------------------------------------
/// Deletes an object. This calls the destructor and frees memory afterwards.
/// Normally you should use DeleteObj(). This variant should be used only in the rare case
/// when the object pointer is a constant so that it can't be set to nullptr.
/// @param[in] o						object pointer (can be nullptr)
/// THREADSAFE
//----------------------------------------------------------------------------------------
template <typename T> void DeleteConstPtrObj(T* o)
{
	if (o)
	{
		o->~T();
		DefaultAllocator::Free(o);
	}
}

//----------------------------------------------------------------------------------------
/// Allocates raw memory: no constructors are called!
/// @return									pointer to memory block or nullptr
/// @param[in] T						type of item (e.g. Char, Float)
/// @param[in] cnt					number of items
/// THREADSAFE
//----------------------------------------------------------------------------------------
#define NewMem(T, cnt) ((T*) c4d_misc::DefaultAllocator::Alloc((c4d_misc::Int64)SIZEOF(T) * (c4d_misc::Int64)cnt, C4D_MISC_ALLOC_LOCATION))

//----------------------------------------------------------------------------------------
/// Allocates raw memory and clears it: no constructors are called!
/// @return									pointer to memory block or nullptr
/// @param[in] T						type of item (e.g. Char, Float)
/// @param[in] cnt					number of items
/// THREADSAFE
//----------------------------------------------------------------------------------------
#define NewMemClear(T, cnt) ((T*) c4d_misc::DefaultAllocator::AllocClear((c4d_misc::Int64)SIZEOF(T) * (c4d_misc::Int64)cnt, C4D_MISC_ALLOC_LOCATION))

//----------------------------------------------------------------------------------------
/// Resizes a raw memory block: no constructors are called!
/// @return									pointer to memory block or nullptr
/// @param[in] p						previous memory block (can be nullptr)
/// @param[in] cnt					new number of items (size of the memory block is n * sizeof(T))
/// THREADSAFE
//----------------------------------------------------------------------------------------
template <typename T> inline T* _ReallocMem(T* p, Int cnt, C4D_MISC_ALLOC_LOCATION_DECLARATION)
{
	Int64	size = SIZEOF(T) * Int64(cnt);

#ifndef __C4D_64BIT
	if (size > c4d_misc::MAXVALUE_INT32)											// overflow?
	{
		DebugStop();
		return nullptr;
	}
#endif

	return (T*) DefaultAllocator::Realloc(p, Int(size), C4D_MISC_ALLOC_LOCATION_FORWARD);
}

//----------------------------------------------------------------------------------------
/// Resizes a raw memory block: no constructors are called!
/// @return									pointer to memory block or nullptr
/// @param[in] p						previous memory block (can be nullptr)
/// @param[in] cnt					new number of items (size of the memory block is n * sizeof(T))
/// THREADSAFE
//----------------------------------------------------------------------------------------
#define ReallocMem(p, cnt) _ReallocMem(p, cnt, C4D_MISC_ALLOC_LOCATION)

//----------------------------------------------------------------------------------------
/// Frees a raw memory block: no destructors are called!
/// @return									-
/// @param[in] p						memory block (can be nullptr, will be nullptr after return)
/// THREADSAFE
//----------------------------------------------------------------------------------------
#define DeleteMem(p) (c4d_misc::DefaultAllocator::Free(p))

} // C4D_MISC_END

#endif
